import numpy as np
import math
import cmath
from scipy import linalg
from numpy.random import normal
from random import choice


def gen_sparse_measurement(n, l, k, delta, p, var_w):
    """
    Generate the measurements according to the MMV model: y = ax + w.

    The measurement matrix a is generated by sparse array idea.
    The signals of interest x (size n*l) is of joint sparsity k and the support satisfies delta-separation.
    [A1] The source signals x are zero-mean and statistically uncorrelated.
    [A2] The additive noise is uncorrelated with signal and zero-mean white.
    Under assumptions [A1] and [A2], for simulations, nonzero elements of x are generated from a multivariate normal
    distribution and the noise term w is also generated independently from a normal distribution.

    Inputs:
        m: number of measurements
        n: signal length
        l: number of snapshots
        k: sparsity
        delta: separation
        p: nonzero signal power vector (if a scalar is provided, it represents equal power signal)
        var_w: variance of the noise term

    Outputs:
        a: measurement matrix
        y: measurements
        x: signals of interest
        r: real form of vectorization of sample covariance of y
        a_tilde: real form of K–R product of measurement matrix
        support: true support of signals of interest
    """

    # Generate the distance map
    theta = math.floor((-1 + math.sqrt(1 + 4*n))/2)
    gamma = 1 + n - theta**2
    m = gamma + theta - 1
    a_map = np.zeros((m, ))

    for index1 in range(gamma):
        a_map[index1] = index1 + 1

    for index1 in range(gamma, m):
        index2 = (index1 + 1 - gamma) * theta + index1 + 1
        a_map[index1] = index2

    # Generate the measurement matrix a
    a = np.zeros((m, n), dtype=complex)
    for i in range(m):
        for j in range(n):
            a[i, j] = cmath.exp(2j * math.pi * a_map[i] * j / n)

    # Generate the noise w
    w = np.random.normal(loc=0, scale=np.sqrt(2)/2*np.sqrt(var_w), size=(m, 2*l)).view(np.complex128)

    # Generate the delta-separated sparse signals x
    x = np.zeros((n, l), dtype=complex)
    candidate = np.arange(n)
    support = np.empty((k,), dtype=int)

    itr = 0
    while itr < k:
        support[itr] = choice(candidate)
        if support[itr] < delta:
            banded_set = np.union1d(np.arange(0, support[itr]+delta+1), np.arange(support[itr]+n-delta, n))
        elif support[itr] > n - delta - 1:
            banded_set = np.union1d(np.arange(support[itr]-delta, n), np.arange(0, support[itr]-n+delta+1))
        else:
            banded_set = np.arange(support[itr]-delta, support[itr]+delta+1)
        candidate = np.setdiff1d(candidate, banded_set)
        itr = itr + 1
    support = np.sort(support)

    if np.size(p) == 1:
        x[support, :] = np.random.normal(loc=0, scale=np.sqrt(2)/2*np.sqrt(p), size=(k, 2*l)).view(np.complex128)
    else:
        for i in range(np.size(p)):
            x[support[i], :] = np.random.normal(loc=0, scale=np.sqrt(2)/2*np.sqrt(p[i]), size=(1, 2*l)).view(np.complex128)

    # Generate the measurements y
    y = a.dot(x) + w

    # Generate the real form of K–R product of measurement matrix a
    a_kha = linalg.khatri_rao(a.conj(), a)
    a_tilde = np.vstack((np.real(a_kha), np.imag(a_kha)))

    # Generate the real form of vectorization of sample covariance y
    y_cov = np.reshape(1 / l * y.dot(y.conj().T), -1, order='F')
    r = np.hstack((np.real(y_cov), np.imag(y_cov)))

    return a, y, x, a_tilde, r, support
