import numpy as np
from gurobipy import Model, GRB, QuadExpr, LinExpr


# Use commercial solver gurobi to solve R_tilde
def perspective(a, y, k, delta, u, lb, ub):
    # Set the optimization model
    model = Model()
    m = a.shape[0]
    n = a.shape[1]
    eig_vals, _ = np.linalg.eig(np.dot(np.transpose(a), a))
    # Define c for numerical error
    c = 1e-6
    min_eig = np.sort(eig_vals)[0] - c

    x = {}
    z = {}
    s = {}

    for column_index in range(n):
        x[column_index] = model.addVar(vtype=GRB.CONTINUOUS, name='x'+str(column_index), ub=u, lb=0)
        z[column_index] = model.addVar(vtype=GRB.CONTINUOUS, name='z'+str(column_index),
                                       ub=ub[column_index], lb=lb[column_index])
        s[column_index] = model.addVar(vtype=GRB.CONTINUOUS, name='s'+str(column_index), ub=GRB.INFINITY, lb=0)

    model.update()

    # Objective
    obj = QuadExpr()
    c = 1/2 * np.full((n, ), min_eig)
    q = 1/2 * np.dot(np.transpose(a), a) - np.diag(c)
    r = - np.dot(np.transpose(y), a)
    for i in range(n):
        for j in range(n):
            if q[i][j] != 0:
                obj += q[i][j] * x[i] * x[j]
    for j in range(n):
        if c[j] != 0:
            obj += c[j] * s[j]
    for j in range(n):
        if r[j] != 0:
            obj += r[j] * x[j]
    obj += 1/2 * np.dot(np.transpose(y), y)
    model.setObjective(obj, GRB.MINIMIZE)

    # Constraints
    for column_index in range(n):
        model.addConstr(x[column_index] <= z[column_index] * u)
        model.addConstr(x[column_index] >= 0)

    for column_index in range(n):
        model.addConstr(x[column_index]**2 <= s[column_index] * z[column_index])

    expr2 = LinExpr()
    for column_index in range(n):
        expr2.addTerms(1.0, z[column_index])
    model.addConstr(expr2 <= k)

    for column_index in range(n-delta):
        expr3 = LinExpr()
        expr3.addTerms([1.0 for _ in range(delta + 1)],
                       [z[index] for index in range(column_index, column_index + delta + 1)])
        model.addConstr(expr3 <= 1)

    for column_index in range(n-delta, n):
        expr4 = LinExpr()
        expr4.addTerms([1.0 for _ in range(column_index - n + delta + 1)],
                       [z[index] for index in range(column_index - n + delta + 1)])
        expr4.addTerms([1.0 for _ in range(n - column_index)],
                       [z[index] for index in range(column_index, n)])
        model.addConstr(expr4 <= 1)

    # model.params.NonConvex = 2
    model.update()
    model.optimize()

    output_x = np.zeros(len(x))
    output_z = np.zeros(len(z))

    for i in range(len(x)):
        output_x[i] = x[i].x
        output_z[i] = z[i].x

    return output_x, output_z, model.ObjVal

