import numpy as np
from Tree import BBTree


def sep_bb(a, y, k, delta, u, lower_solver='cvx', gap_tol=1e-2, time_limit=1800):
    """
    Solve the separation-based l0-constrained least square problem through Sep-B&B.

    Inputs:
        a: measurement matrix
        y: measurements
        k: maximum sparsity
        u: element-wise upper bound
        delta: separation
        lower_solver: 'cvx' or 'perspective' or 'gurobi' (how to solve the relaxation problem)
        gap_tol: numerical gap tolerance of branch-and-bound algorithm
        time_limit: maximum running time

    Outputs:
        num_of_nodes: number of visited nodes
        support: recovered support
        time: recovery time
    """

    # Centralize and normalize measurements y and measurement matrix a
    a_centered, y_centered = preprocess(a, y)

    # Construct a BB tree
    tree = BBTree(a_centered, y_centered)

    # Call Sep-B&B to solve the problem (Co-MIP)
    tree_sol = tree.solve(k, delta, u, lower_solver, gap_tol, time_limit)

    # Results
    num_of_nodes = tree_sol.num_of_nodes
    support = tree_sol.support
    time = tree_sol.sol_time

    return num_of_nodes, support, time


def preprocess(a, y):
    """
    Before we construct the BnB tree, we first centralize and normalize y and a.
    Centralize: make measurement y and every column of the measurement matrix a zero-centered
    Normalize: make l2 norm of measurement y and every column of the measurement matrix a to be one

    Inputs:
        a: measurement matrix
        y: measurements

    Outputs:
        a_centered: a after centralization and normalization
        y_centered: y after centralization and normalization
    """

    # Centralize
    a_mean = np.mean(a, axis=0)
    a_centered = a - a_mean
    y_mean = np.mean(y)
    y_centered = y - y_mean

    # Normalize
    a_centered_norm = np.linalg.norm(a_centered, axis=0)
    y_centered_norm = np.linalg.norm(y_centered)
    a_centered = a_centered / a_centered_norm
    y_centered = y_centered / y_centered_norm

    return a_centered, y_centered
