import numpy as np
import math
import time
from scipy.optimize import linprog


def sep_cosamp(a, y, k, delta, u, itr_max=10):
    """
    Solve Sep-CoSaMP to approximate the support of power vector or provide an upper bound of Sep-B&B

    Inputs:
    a: measurement matrix
    y: measurements
    k: sparsity
    u: element-wise upper bound
    delta: separation
    itr_max: max iteration (for CoSaMP, 10 is enough in literature)

    Outputs:
    upper_bound_cosamp: objective value under x_cosamp
    support_cosamp: approximate support of power vector x_cosamp under model-based CoSaMP
    x_cosamp: approximate power vector under model-based CoSaMP
    """

    # Initialization
    start = time.time()
    n = a.shape[1]
    x = np.empty((itr_max,), dtype=object)
    r = np.empty((itr_max,), dtype=object)
    support = np.empty((itr_max,), dtype=object)
    itr = 0

    # Sep-CoSaMP
    while itr < itr_max:
        if itr == 0:
            e = np.dot(a.conj().T, y)
        else:
            e = np.dot(a.conj().T, r[itr-1])
        omega, _ = structure_estimate(e, delta, 2, 2*k)
        if itr == 0:
            t = np.sort(omega)
        else:
            t = np.sort(np.union1d(omega, support[itr-1]))
        b = np.zeros((n, ))
        b[t] = np.dot(np.linalg.pinv(a[:, t]), y)
        support[itr], x[itr] = structure_estimate(b, delta, 1, k)
        x[itr][x[itr] > u] = u
        r[itr] = y - np.dot(a, x[itr])
        itr = itr+1
    support_cosamp = support[itr_max-1]
    x_cosamp = x[itr_max-1]
    upper_bound_cosamp = 1/2*math.pow(np.linalg.norm(r[itr_max-1]), 2)

    return upper_bound_cosamp, x_cosamp, support_cosamp, time.time()-start


def structure_estimate(x, delta, r, k):
    """
    Project x under the separation prior.

    Inputs:
    x: signal to be projected
    delta: separation
    r: specified number of spikes
    k: specified number of sparsity

    Outputs:
    support_est: approximate support of x under structure estimation
    x_est: approximate x under structure estimation
    """

    # Initialization
    n = x.size
    # c = np.power(abs(x), 2)
    c = x
    w = np.zeros((n+1, n))
    w[0, :] = np.ones((1, n))
    for i in range(0, n-delta):
        w[i+1, i:(i+delta+1)] = 1
    for i in range(n-delta, n):
        w[i+1, i:n] = 1
        w[i+1, 0:(i-n+delta+1)] = 1
    b = np.zeros((n+1,))
    b[0] = k
    b[1:n+1] = r
    res = linprog(c=-c, A_ub=w, b_ub=b, bounds=(0, 1), method='revised simplex')
    res.x[res.x < math.pow(10, -4)] = 0
    res.x[res.x > 1-math.pow(10, -4)] = 1
    x_est = np.multiply(x, res.x)
    support_est = np.sort(np.nonzero(res.x > math.pow(10, -4))).flatten()

    return support_est, x_est
