import time
import queue
import sys
import numpy as np
from collections import namedtuple
from Sep_CoSaMP import sep_cosamp
from Node import BBNode
from Branching import is_integral, branch


class BBTree:
    def __init__(self, a, y, int_tol=1e-6):
        """
        Construct a BB tree to solve the separation-based l0-constrained least square problem.

        Inputs:
        a: measurement matrix
        y: measurements
        int_tol: numerical integer tolerance of solutions of continuous relaxation
        """

        self.a = a
        self.y = y
        self.int_tol = int_tol

        self.m = a.shape[0]
        self.n = a.shape[1]

        self.dfs_queue = None

        self.levels = {}
        self.num_of_nodes = 0

        self.root = None

    def solve(self, k, delta, u, lower_solver, gap_tol, time_limit):
        """
        Solve the separation-based l0-constrained least square problem.

        Inputs:
        k: maximum sparsity
        u: element-wise upper bound
        delta: separation
        gap_tol: numerical gap tolerance of branch-and-bound algorithm which determines the optimal solution
        lower_solver:
        time_limit: maximum running time

        Outputs:
        A solution tuple including recovered support, recovery time and the number of visited nodes
        """

        start = time.time()

        # Initialize the upper bound by model-based CoSaMP or warm start
        upper_bound, upper_x, support = self._warm_start(k, delta, u)

        # Define the root node and the queue
        self.root = BBNode(None, zlb=[], zub=[], a=self.a, y=self.y)
        self.dfs_queue = queue.LifoQueue()
        self.dfs_queue.put(self.root)

        # Initialize the lower bound and the gap
        lower_bound = {}
        self.levels = {0: 1}
        min_open_level = 0
        max_lower_bound_value = -sys.maxsize
        best_gap = 1

        while self.dfs_queue.qsize() > 0 and (time.time() - start < time_limit):
            # Push a node (denoted by current_node)
            current_node = self.dfs_queue.get()

            # Check whether to prune?
            if current_node.parent_lower_bound and upper_bound <= current_node.parent_lower_bound:
                self.levels[current_node.level] -= 1
                continue

            # Calculate the lower bound of current_node (denoted by current_lower_bound)
            current_lower_bound = self._solve_relaxation(current_node, k, delta, u, lower_solver, lower_bound)

            # Update gap
            if self.levels[min_open_level] == 0:
                del self.levels[min_open_level]
                max_lower_bound_value = max([j for i, j in lower_bound.items() if i <= min_open_level])
                best_gap = (upper_bound - max_lower_bound_value) / abs(upper_bound)
                min_open_level += 1

            # Check whether we have arrived at a solution
            if best_gap <= gap_tol:
                return self._package_solution(support, self.num_of_nodes, time.time()-start)

            # Integral solution with sparsity k?
            if is_integral(current_node.z, current_node.lower_x, self.int_tol):
                current_upper_bound = current_lower_bound
                if current_upper_bound < upper_bound:
                    upper_bound = current_upper_bound
                    upper_x = current_node.upper_x
                    support = current_node.support
                best_gap = (upper_bound - max_lower_bound_value) / abs(upper_bound)
            # Branch?
            elif current_lower_bound < upper_bound:
                left_node, right_node = branch(current_node, self.int_tol, delta)
                self.levels[current_node.level + 1] = self.levels.get(current_node.level + 1, 0) + 2
                self.dfs_queue.put(right_node)
                self.dfs_queue.put(left_node)
            else:
                pass
        return self._package_solution(support, self.num_of_nodes, time.time() - start)

    def _warm_start(self, k, delta, u):
        upper_bound, upper_x, support, _ = sep_cosamp(self.a, self.y, k, delta, u)
        return upper_bound, upper_x, support

    def _solve_relaxation(self, current_node, k, delta, u, lower_solver, lower_bound):
        self.num_of_nodes += 1
        current_lower_bound = current_node.lower_solve(k, delta, u, lower_solver, self.int_tol)
        lower_bound[current_node.level] = min(current_lower_bound, lower_bound.get(current_node.level, sys.maxsize))
        self.levels[current_node.level] -= 1
        return current_lower_bound

    @staticmethod
    def _package_solution(support, num_of_nodes, sol_time):
        _sol_str = 'sol_time support num_of_nodes'
        Solution = namedtuple('Solution', _sol_str)
        return Solution(sol_time=sol_time, support=np.array(support), num_of_nodes=num_of_nodes)
